/*
** Copyright (C) 2002-2003 Ryan McCabe <ryan@numb.org>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License, version 2,
** as published by the Free Software Foundation.
*/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <time.h>
#include <pwd.h>
#include <grp.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/stat.h>
#include <sys/signal.h>

#include <pthread.h>

#include <snuff.h>
#include <snuff_list.h>
#include <snuff_util.h>
#include <snuff_db.h>

static void print_help_text(void);
static void print_version_info(void);
static int setup_listen_sock(void);
static void *connection_handler(void *d);
static void server_shutdown(int s);
static void start_expire_thread(void);
static int get_creds(int sock, struct ucred *ucred);
static void seed_random(void);

static u_int32_t max_len = DEFAULT_USERLEN + 1;
u_int32_t max_aliases = 0;

int main(int argc, char **argv) {
	int opt;
	int sock;
	uid_t uid = getuid();
	gid_t gid = getgid();

	seed_random();

	while ((opt = getopt(argc, argv, "u:g:hvl:m:")) != EOF) {
		switch (opt) {
			case 'l':
			{
				char *end;

				max_len = strtoul(optarg, &end, 10) + 1;
				if (*end != '\0') {
					fprintf(stderr, "Error: Bad length: \"%s\"\n", optarg);
					exit(-1);
				}

				if (max_len > MAX_USERLEN) {
					fprintf(stderr,
						"The maximum length must be less than %u\n",
						MAX_USERLEN);
					exit(-1);
				}
				break;
			}

			case 'm':
			{
				char *end;

				max_aliases = strtoul(optarg, &end, 10);
				if (*end != '\0') {
					fprintf(stderr, "Bad number: %s\n", optarg);
					exit(-1);
				}

				break;

			}

			case 'u':
				if (find_user(optarg, &uid) != 0) {
					fprintf(stderr, "Invalid user: \"%s\"\n", optarg);
					exit(-1);
				}
				break;

			case 'g':
				if (find_group(optarg, &gid) != 0) {
					fprintf(stderr, "Invalid group: \"%s\"\n", optarg);
					exit(-1);
				}
				break;

			case 'v':
				print_version_info();
				exit(0);
				break;

			case 'h':
			default:
				print_help_text();
				exit(0);
				break;
		}
	}

	sock = setup_listen_sock();
	if (sock == -1) {
		fprintf(stderr, "Unable to setup listen sock.\n");
		exit(-1);
	}

	if (drop_privs(uid, gid) != 0) {
		fprintf(stderr, "Unable to drop privileges.\n");
		exit(-1);
	}

	if (go_background() != 0) {
		fprintf(stderr, "Unable to drop privileges.\n");
		exit(-1);
	}

	if (init_state() == -1) {
		fprintf(stderr, "Unable to initialize the state.\n");
		exit(-1);
	}

	read_state(STATE_FILE);

	signal(SIGPIPE, SIG_IGN);
	signal(SIGTERM, server_shutdown);
	signal(SIGINT, server_shutdown);

	start_expire_thread();

	while (1) {
		struct sockaddr_un peer_addr;
		socklen_t len;
		int peer;
		pthread_t pt;
		pthread_attr_t thread_attr;

		len = sizeof(peer_addr);

		peer = accept(sock, (struct sockaddr *) &peer_addr, &len);
		if (peer == -1) {
			debug("Error: accept: %s", strerror(errno));
			continue;
		}

		pthread_attr_init(&thread_attr);
		pthread_attr_setdetachstate(&thread_attr,
			PTHREAD_CREATE_DETACHED);

		pthread_create(&pt, &thread_attr, connection_handler, (void *) peer);
	}

	exit(0);
}

static int setup_listen_sock(void) {
	struct sockaddr_un sun;
	int sock;

	/* Doesn't matter if this fails. */
	unlink(MSOCK_PATH);

	sock = socket(PF_UNIX, SOCK_STREAM, 0);
	if (sock == -1) {
		debug("Error: socket: %s", strerror(errno));
		return (-1);
	}

	sun.sun_family = AF_UNIX;

	if (strlen(MSOCK_PATH) >= sizeof(sun.sun_path)) {
		debug("Error: socket path is too long.");
		close(sock);
		return (-1);
	}

	strcpy(sun.sun_path, MSOCK_PATH);

	if (bind(sock, (struct sockaddr *) &sun, sizeof(sun)) != 0) {
		debug("Error: bind: %s", strerror(errno));
		close(sock);
		return (-1);
	}

	/*
	** Make it such that everyone can connect.
	*/

	if (chmod(MSOCK_PATH, 0666) != 0) {
		debug("Error: chmod: %s\n", strerror(errno));
		close(sock);
		unlink(MSOCK_PATH);
		return (-1);
	}

	if (listen(sock, 3) != 0) {
		debug("Error: listen: %s", strerror(errno));
		close(sock);
		unlink(MSOCK_PATH);
		return (-1);
	}

	return (sock);
}

static int get_creds(int sock, struct ucred *ucred) {
	int one = 1;
	struct iovec iov;
	struct msghdr msg;
	char data[sizeof(struct cmsghdr) + sizeof(struct ucred)];
	struct cmsghdr *cmsg;
	char buf;

	if (setsockopt(sock, SOL_SOCKET, SO_PASSCRED, &one, sizeof(one)) == -1) {
		debug("Error: setsockopt: %s", strerror(errno));
		return (-1);
	}

	/*
	** For some reason, it doesn't like it when I pass in a NULL
	** for the base and a 0 for the len..
	*/

	iov.iov_base = &buf;
	iov.iov_len = sizeof(buf);

	memset(&msg, 0, sizeof(msg));

	msg.msg_iov = &iov;
	msg.msg_iovlen = 1;
	msg.msg_control = data;
	msg.msg_controllen = sizeof(data);

	if (recvmsg(sock, &msg, MSG_PEEK) == -1 || msg.msg_flags != 0) {
		debug("Error: recvmsg: %s", strerror(errno));
		return (-1);
	}

	cmsg = CMSG_FIRSTHDR(&msg);

	while (cmsg != NULL) {
		if (cmsg->cmsg_level == SOL_SOCKET &&
			cmsg->cmsg_type == SCM_CREDENTIALS)
		{
			memcpy(ucred, CMSG_DATA(cmsg), sizeof(*ucred));
			return (0);
		}

		cmsg = CMSG_NXTHDR(&msg, cmsg);
	}

	return (-1);
}

static inline int get_request(int sock, struct control_block *request) {
	int ret;

	ret = sock_read(sock, request, sizeof(*request));
	if (ret != sizeof(*request)) {
		debug("Error: read: %s", strerror(errno));
		return (-1);
	}

	return (0);
}

static inline int send_result(int sock, int result_code) {
	int ret;

	ret = sock_write(sock, &result_code, sizeof(result_code));
	if (ret != sizeof(result_code)) {
		debug("Error: write: %s", strerror(errno));
		return (-1);
	}

	return (0);
}

static void *connection_handler(void *d) {
	sigset_t csigset;
	int sock = (int) d;
	struct ucred ucred;
	struct control_block request;
	int ret;

	sigemptyset(&csigset);
	sigaddset(&csigset, SIGINT);
	sigaddset(&csigset, SIGTERM);
	sigaddset(&csigset, SIGHUP);
	sigaddset(&csigset, SIGPIPE);
	sigaddset(&csigset, SIGQUIT);

	pthread_sigmask(SIG_BLOCK, &csigset, NULL);

	if (get_creds(sock, &ucred) == -1) {
		debug("Unable to verify client's credentials");
		goto out;
	}

	if (get_request(sock, &request) == -1) {
		debug("Unable to read request");
		goto out;
	}

	switch (request.action) {
		case ACTION_ADD:
		{
			struct alias_block alias;

			alias.expire = request.expire;

			ret = new_alias(ucred.uid, &alias, max_len);
			if (ret != 0) {
				debug("Error adding alias");
				send_result(sock, ret);
				goto out;
			}

			send_result(sock, RESULT_SUCCESS);

			ret = sock_write(sock, alias.username, sizeof(alias.username));
			if (ret != sizeof(alias.username)) {
				debug("Error: write: %s", strerror(errno));
				goto out;
			}

			break;
		}

		case ACTION_DELETE:
			ret = delete_alias(ucred.uid, request.username);
			send_result(sock, ret);
			break;

		case ACTION_DELETE_ALL:
			ret = delete_all_aliases(ucred.uid);
			send_result(sock, ret);
			break;

		case ACTION_LIST:
			send_result(sock, RESULT_SUCCESS);
			list_aliases(sock, ucred.uid);
			break;

		default:
			send_result(sock, RESULT_INVALID_ACTION);
			goto out;
	}

out:
	close(sock);
	pthread_exit(NULL);
}

static void *expire_thread(void *unused) {
	sigset_t csigset;

	sigemptyset(&csigset);
	sigaddset(&csigset, SIGINT);
	sigaddset(&csigset, SIGTERM);
	sigaddset(&csigset, SIGHUP);
	sigaddset(&csigset, SIGQUIT);

	pthread_sigmask(SIG_BLOCK, &csigset, NULL);

	while (1) {
		expire_aliases();
		xsleep(1);
	}

	return (unused);
}

static void start_expire_thread(void) {
	pthread_attr_t thread_attr;
	pthread_t pt;

	pthread_attr_init(&thread_attr);
	pthread_attr_setdetachstate(&thread_attr,
		PTHREAD_CREATE_DETACHED);

	pthread_create(&pt, &thread_attr, expire_thread, NULL);

}

static void server_shutdown(int sig) {
	(void) sig;

	save_state(STATE_FILE);
	unlink(MSOCK_PATH);
	exit(0);
}

static void seed_random(void) {
	struct timeval tv;

	gettimeofday(&tv, NULL);

	srandom(tv.tv_sec ^ (tv.tv_usec << 11));
}

static void print_help_text(void) {
	fprintf(stderr,
"Usage: snuffd [options]\n\
-l <num>  Maximum length of aliases\n\
-m <num>  Maximum number of concurrent aliases users are allowed to have (defaults to unlimited)\n\
-u        Run as the specified user or UID\n\
-g        Run as with the specified group or GID\n\
-h        Print this help text\n\
-v        Print version information\n");
}

static void print_version_info(void) {
	printf("snuffd version %s by %s\n", SNUFF_VERSION, SNUFF_AUTHOR);
}
